@@ -19,8 +19,8 @@ fn case() {
         .current_dir(project_root)
         .assert()
         .success()
-        .stdout_matches(str![""])
-        .stderr_matches(file!["stderr.term.svg"]);
+        .stdout_eq(str![""])
+        .stderr_eq(file!["stderr.term.svg"]);
 
     assert_ui().subset_matches(current_dir!().join("out"), project_root);
     assert!(!project_root.join(".gitignore").is_file());
