@@ -1,5 +1,4 @@
-// compile-flags: -O
-// min-llvm-version: 16
+//@ compile-flags: -O
 
 #![crate_type = "lib"]
 
@@ -10,9 +9,5 @@ pub fn test(a: i32, b: i32) -> bool {
     let c1 = (a >= 0) && (a <= 10);
     let c2 = (b >= 0) && (b <= 20);
 
-    if c1 & c2 {
-        a + 100 != b
-    } else {
-        true
-    }
+    if c1 & c2 { a + 100 != b } else { true }
 }
