@@ -6,17 +6,14 @@
 // to prevent LLVM from inferring the attribute.
 
 #![crate_type = "lib"]
-#![feature(c_unwind)]
 
 // CHECK: @rust_item_that_cannot_unwind() unnamed_addr #0 {
 #[no_mangle]
-pub extern "C" fn rust_item_that_cannot_unwind() {
-}
+pub extern "C" fn rust_item_that_cannot_unwind() {}
 
 // CHECK: @rust_item_that_can_unwind() unnamed_addr #1 {
 #[no_mangle]
-pub extern "C-unwind" fn rust_item_that_can_unwind() {
-}
+pub extern "C-unwind" fn rust_item_that_can_unwind() {}
 
 // Now, make some assertions that the LLVM attributes for these functions are correct.  First, make
 // sure that the first item is correctly marked with the `nounwind` attribute:
