@@ -1,5 +1,6 @@
-// compile-flags: -Zunleash-the-miri-inside-of-you
-// stderr-per-bitwidth
+//@ compile-flags: -Zunleash-the-miri-inside-of-you
+//@ normalize-stderr: "(the raw bytes of the constant) \(size: [0-9]*, align: [0-9]*\)" -> "$1 (size: $$SIZE, align: $$ALIGN)"
+//@ normalize-stderr: "([0-9a-f][0-9a-f] |╾─*ALLOC[0-9]+(\+[a-z0-9]+)?(<imm>)?─*╼ )+ *│.*" -> "HEX_DUMP"
 
 use std::sync::atomic::AtomicUsize;
 use std::sync::atomic::Ordering;
@@ -17,17 +18,16 @@ const READ_INTERIOR_MUT: usize = {
 static mut MUTABLE: u32 = 0;
 const READ_MUT: u32 = unsafe { MUTABLE }; //~ERROR evaluation of constant value failed
 
-const REF_INTERIOR_MUT: &usize = { //~ ERROR undefined behavior to use this value
-//~| encountered a reference pointing to a static variable
+// Evaluating this does not read anything mutable, but validation does, so this should error.
+const REF_INTERIOR_MUT: &usize = { //~ ERROR undefined behavior
+    //~| encountered reference to mutable memory
     static FOO: AtomicUsize = AtomicUsize::new(0);
     unsafe { &*(&FOO as *const _ as *const usize) }
 };
 
-// ok some day perhaps
-const READ_IMMUT: &usize = { //~ ERROR it is undefined behavior to use this value
-//~| encountered a reference pointing to a static variable
-    static FOO: usize = 0;
-    &FOO
-};
+// Not actually reading from anything mutable, so these are fine.
+static MY_STATIC: u8 = 4;
+const REF_IMMUT: &u8 = &MY_STATIC;
+const READ_IMMUT: u8 = *REF_IMMUT;
 
 fn main() {}
