@@ -1,8 +1,8 @@
 //@ edition:2015
-//@ aux-build:edition-kw-macro-2015.rs
+//@ aux-build:edition-kw-macro-2018.rs
 
 #[macro_use]
-extern crate edition_kw_macro_2015;
+extern crate edition_kw_macro_2018;
 
 mod module {
     pub fn async() {}
@@ -13,8 +13,8 @@ pub fn check_async() {
     let mut r#async = 1; // OK
 
     r#async = consumes_async!(async); // OK
-    r#async = consumes_async!(r#async); //~ ERROR no rules expected the token `r#async`
-    r#async = consumes_async_raw!(async); //~ ERROR no rules expected the token `async`
+    r#async = consumes_async!(r#async); //~ ERROR no rules expected `r#async`
+    r#async = consumes_async_raw!(async); //~ ERROR no rules expected `async`
     r#async = consumes_async_raw!(r#async); // OK
 
     if passes_ident!(async) == 1 {} // OK
