@@ -1,6 +1,15 @@
 #![feature(type_alias_impl_trait)]
 
-type Foo = impl PartialEq<(Foo, i32)>;
+//@ check-pass
+
+mod foo {
+    pub type Foo = impl PartialEq<(Foo, i32)>;
+
+    fn foo() -> Foo {
+        super::Bar
+    }
+}
+use foo::Foo;
 
 struct Bar;
 
@@ -10,9 +19,4 @@ impl PartialEq<(Foo, i32)> for Bar {
     }
 }
 
-fn foo() -> Foo {
-    //~^ ERROR can't compare `Bar` with `(Bar, i32)`
-    Bar
-}
-
 fn main() {}
