@@ -1,6 +1,5 @@
 #![feature(intrinsics)]
 #![feature(rustc_attrs)]
-#![feature(effects)]
 
 extern "rust-intrinsic" {
     fn size_of<T>() -> usize; //~ ERROR intrinsic safety mismatch
@@ -19,7 +18,6 @@ const fn const_deallocate(_ptr: *mut u8, _size: usize, _align: usize) {}
 mod foo {
     #[rustc_intrinsic]
     unsafe fn const_deallocate(_ptr: *mut u8, _size: usize, _align: usize) {}
-    //~^ ERROR wrong number of const parameters
 }
 
 fn main() {}
