@@ -1,6 +1,5 @@
+#![feature(exhaustive_patterns)]
 #![feature(never_type, never_type_fallback)]
-#![feature(min_exhaustive_patterns)]
-
 #![allow(unreachable_code)]
 #![deny(unreachable_patterns)]
 
